#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <semaphore.h>
#include <pthread.h>

sem_t sem_sort;
sem_t sem_sys;
sem_t sem_local;

struct params
{
	int start_index;
	int end_index;
};

int *elements;
int *sorted_elements;

bool done = false;
int no_of_elements;

#define MAXLINE 256

void merge(int partial[], int size)
{
	sem_wait(&sem_sort);

	static bool first_time = true;
	static int current_no_of_elements = 0;

	if(first_time)
	{
		first_time = false;
		memcpy(sorted_elements, partial, size * sizeof(int));

		current_no_of_elements = size;
		sem_post(&sem_sort);
		return;

	}

	int index1 = 0, index2 = 0,index3 = 0;

	int *temp = new int[current_no_of_elements + size];

	while(index1 < current_no_of_elements && index2 < size)
	{
		if(sorted_elements[index1] < partial[index2])
			temp[index3++] = sorted_elements[index1++];
		else if(partial[index2] < sorted_elements[index1])
			temp[index3++] = partial[index2++];
		else
		{
			temp[index3++] = partial[index2++];
			index1++;
		}

	}

	while(index1 < current_no_of_elements)
		temp[index3++] = sorted_elements[index1++];

	while(index2 < size)
		temp[index3++] = partial[index2++];

	current_no_of_elements += size;

	memcpy(sorted_elements, temp, current_no_of_elements * sizeof(int));

	sem_wait(&sem_local);
	if(current_no_of_elements == no_of_elements)
		done = true;
	sem_post(&sem_local);

	delete[] temp;

	sem_post(&sem_sort);
}

void* start_thread(void *arg)
{
	struct params t = *((struct params*) arg);

	int fd[2];
	
	int ret_val;

	sem_wait(&sem_sys);
	ret_val = pipe(fd);
	sem_post(&sem_sys);

	if(ret_val < 0)
		cerr << "pipe error" << endl;
	
	pid_t pid;

	sem_wait(&sem_sys);
	pid = fork();
	sem_post(&sem_sys);

	if(pid < 0)
		cerr << "fork error" << endl;
	else if(pid > 0)
	{
		sem_wait(&sem_sys);
		close(fd[1]);
		sem_post(&sem_sys);

		char buf[MAXLINE];
		int p_sum;
		
		sem_wait(&sem_sys);
		read(fd[0], buf, MAXLINE);
		sem_post(&sem_sys);

		int temp[MAXLINE];

		for(int i = 0; i < t.end_index - t.start_index;i++)
			memcpy(&temp[i], &buf[sizeof(int)*i], sizeof(int));

		merge(temp, t.end_index - t.start_index);

	}
	else
	{
		close(fd[0]);

		int sum = 0;
		
		cout << "\nStart: " << t.start_index <<endl;
		cout << "End: " << t.end_index << endl;

		for(int i = t.start_index; i < t.end_index - 1;i++)
			for(int j = t.start_index;j < t.end_index - 1;j++)
				if(elements[j] > elements[j+1])
				{
					int temp = elements[j];
					elements[j] = elements[j+1];
					elements[j+1] = temp;
				}

		cout << "Partially sorted array: ";

		for(int i = t.start_index;i < t.end_index;i++)
			cout << elements[i] << " ";
		cout << endl;
			
		char buf[MAXLINE];
		
		memcpy(buf, &elements[t.start_index], (t.end_index - t.start_index) * sizeof(int));
		write(fd[1], buf, MAXLINE);
		close(fd[1]);

		exit(0);
		
	}

	pthread_exit(0);
	return 0;
}

void main()
{
	sem_init(&sem_sys, 0, 0);
	sem_post(&sem_sys);

	sem_init(&sem_local, 0, 0);
	sem_post(&sem_local);
	
	sem_init(&sem_sort, 0, 0);
	sem_post(&sem_sort);

	cout << "Enter the no of elements: ";
	cin >> no_of_elements;

	elements = new int[no_of_elements];
	sorted_elements = new int[no_of_elements];
	
	cout << "Enter the elements: " << endl;

	for(int i =0; i < no_of_elements;i++)
		cin >> elements[i];

	int interval = 3;
	int current_index = 0;

	while(current_index < no_of_elements)
	{
		int end_index = ((current_index + interval) < no_of_elements) ? current_index + interval : no_of_elements;
		
		struct params *temp = new params;
		temp->start_index = current_index;
		temp->end_index = end_index;
		
		pthread_t new_thread;
		pthread_create(&new_thread, 0, &start_thread, temp);

		current_index = end_index;
	}
	
	for(;;)
	{
		sem_wait(&sem_local);
		if(done)
		{
			cout << "\nSorted array: ";
			
			for(int i = 0;i < no_of_elements;i++)
				cout << sorted_elements[i] << " ";
			cout << endl;
			delete[] elements;
			delete[] sorted_elements;

			sem_post(&sem_local);
			break;
		}
		sem_post(&sem_local);
	}
}	
