/* TCP Server */

#include <stdio.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#define SERVER_TCP_PORT  3000
#define BUFLEN 		 256

struct thread_params
{
	int new_sd;
	int current_index;
};

sem_t sem_sys;
sem_t sem_local;

int sum = 0;
int interval = 3;
int* elements;
int no_of_elements;
int current_index = 0;

int no_calculated = 0;

void* distribute_problem(void *arg)
{
	struct thread_params *t = (struct thread_params*) arg;

	int new_sd = t->new_sd;
	int ci = t->current_index;

	/*** write the array to the client ***/

	char buf[BUFLEN];
	
	int in;
	
	sem_wait(&sem_local);
	in = interval;
	sem_post(&sem_local);

	for(int i = 0;i < in && ci + i < no_of_elements;i++)
		memcpy(&buf[sizeof(int)*i], &elements[ci + i], sizeof(int));

	int n;	
	for(;;)
	{
		sem_wait(&sem_sys);
		n = write(new_sd, buf, BUFLEN);
		sem_post(&sem_sys);
	
		if(n <= 0)
		{
			if(errno != EAGAIN)
			{
				cerr << "write failed" << endl;		
				pthread_exit(arg);
			}
		}
		else
			break;
	}

	/*** read th partial sum ***/

	char *bp = buf;
	int bytes_to_read = BUFLEN;

	for(;;)
	{
		sem_wait(&sem_sys);
		n = read(new_sd, bp, bytes_to_read);
		sem_post(&sem_sys);

		if(n <= 0)
		{
			if(errno != EAGAIN)
			{
				cerr << "read failed" << endl;	
				pthread_exit(arg);
			}
			else
				continue;
		}

		bytes_to_read -= n;
		bp += n;

		if(!bytes_to_read)
			break;
				
	}

	int partial_sum;

	memcpy(&partial_sum, buf, sizeof(int));

	cout << "Partial sum: " << partial_sum << endl;

	sem_wait(&sem_local);
	sum += partial_sum;
	sem_post(&sem_local);

	sem_wait(&sem_sys);
	close(new_sd);
	sem_post(&sem_sys);

	sem_wait(&sem_local);
	no_calculated += interval;
	sem_post(&sem_local);	
	
	pthread_exit(0);
	return 0;
	
}

int main(int argc, char** argv)
{
	sem_init(&sem_sys, 0, 0);
	sem_post(&sem_sys);
	
	sem_init(&sem_local, 0, 0);
	sem_post(&sem_local);

	int n;
	int bytes_to_read;
	int port;
	int sd, new_sd;
	socklen_t client_len;

	struct sockaddr_in server, client;
	char *bp;
	char buf[BUFLEN];

	cout << "\nEnter the number of elements: ";	
	cin >> no_of_elements;

	elements = new int[no_of_elements];
	cout << "Enter the elements: " << endl;

	for(int i = 0;i < no_of_elements;i++)
	{
		cout << i+1 << ": ";
		cin >> elements[i];
	}

	current_index = 0;

	port = SERVER_TCP_PORT;

	/* create a stream socket */

	if((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		cerr << "server: can't create a socket";
		exit(1);
	}

	/* Bind address to a socket */

	bzero((char*) &server, sizeof(struct sockaddr_in));
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	server.sin_addr.s_addr = htonl(INADDR_ANY);

	if(bind(sd, (struct sockaddr*) &server, sizeof(server)) == -1)
	{
		cerr << "can't bind name to socket";
		exit(1);
	}

	listen(sd, 5);

	while(1)
	{
		client_len = sizeof(client);

		if((new_sd = accept(sd, (struct sockaddr*) &client, &client_len)) == -1)
		{
			cerr << "server: can't accept client\n";
			exit(1);
		}

		if(fcntl(new_sd, F_SETFL, O_NONBLOCK) == -1)
		{
			cerr << "fcntl failed" << endl;
			exit(1);
		}

		pthread_t new_thread;
		struct thread_params t;
		t.new_sd = new_sd;
		t.current_index = current_index;

		pthread_create(&new_thread, 0, &distribute_problem ,&t);
		
		sem_wait(&sem_local);
		current_index = current_index + interval < no_of_elements ? current_index + interval : no_of_elements ;
		sem_post(&sem_local);

		if ( current_index == no_of_elements)
			break;
	}
	
	sem_wait(&sem_sys);
	close(sd);
	sem_post(&sem_sys);

	for(;;)
	{
		sem_wait(&sem_sys);
		if(no_calculated >= no_of_elements)
		{
			cout << "Sum = " << sum << endl;
			sem_post(&sem_sys);
			break;
		}
		sem_post(&sem_sys);

	}
	return 0;
}
