/* TCP Server */

#include <stdio.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#define SERVER_TCP_PORT  3000
#define BUFLEN 		 256

struct thread_params
{
	int new_sd;
	int current_index;
};

sem_t sem_sys;
sem_t sem_local;
sem_t sem_sort;

int interval = 3;
int* elements;
int* sorted_elements;
int no_of_elements;
int current_index = 0;

bool done = false;

void merge(int partial[], int size)
{
	sem_wait(&sem_sort);
	
	static bool first_time = true;
	static int current_no_of_elements = 0;

	if(first_time)
	{
		first_time = false;
		memcpy(sorted_elements, partial, size * sizeof(int));
			
		current_no_of_elements = size;
		sem_post(&sem_sort);
		return;
		
	}

	int index1 = 0, index2 = 0,index3 = 0;

	int *temp = new int[current_no_of_elements + size];
	
	while(index1 < current_no_of_elements && index2 < size)
	{
		if(sorted_elements[index1] < partial[index2])
			temp[index3++] = sorted_elements[index1++];
		else if(partial[index2] < sorted_elements[index1])
			temp[index3++] = partial[index2++];
		else
		{
			temp[index3++] = partial[index2++];
			index1++;
		}
					
	}

	while(index1 < current_no_of_elements)
		temp[index3++] = sorted_elements[index1++];

	while(index2 < size)
		temp[index3++] = partial[index2++];

	current_no_of_elements += size;
	
	memcpy(sorted_elements, temp, current_no_of_elements * sizeof(int));
	
	delete[] temp;

	sem_wait(&sem_local);
	if(current_no_of_elements == no_of_elements)
		done = true;
	sem_post(&sem_local);
	
	sem_post(&sem_sort);	
}

void* distribute_problem(void *arg)
{
	struct thread_params *t = (struct thread_params*) arg;

	int new_sd = t->new_sd;
	int ci = t->current_index;

	/*** write the array to the client ***/

	char buf[BUFLEN];
	
	int in, noe;
	
	sem_wait(&sem_local);
	in = interval;
	noe = no_of_elements;
	sem_post(&sem_local);

	int size = ci + in < noe ? in : no_of_elements - ci;

	cout << "size: " << size << endl;
	/*** write number of elements first ***/
	memcpy(buf, &size, sizeof(int));

	for(int i = 0;i < size;i++)
		memcpy(&buf[sizeof(int)*(i+1)], &elements[ci + i], sizeof(int));

	int n;	
	for(;;)
	{
		sem_wait(&sem_sys);
		n = write(new_sd, buf, BUFLEN);
		sem_post(&sem_sys);
	
		if(n <= 0)
			{
			if(errno != EAGAIN)
			{
				cerr << "write failed" << endl;		
				pthread_exit(arg);
			}
		}
		else
			break;
	}

	/*** read the partially sorted array ***/

	char *bp = buf;
	int bytes_to_read = BUFLEN;

	for(;;)
	{
		sem_wait(&sem_sys);
		n = read(new_sd, bp, bytes_to_read);
		sem_post(&sem_sys);

		if(n <= 0)
		{
			if(errno != EAGAIN)
			{
				cerr << "read failed" << endl;	
				pthread_exit(arg);
			}
			else
				continue;
		}

		bytes_to_read -= n;
		bp += n;

		if(!bytes_to_read)
			break;
				
	}


	int *temp = new int[size];

	for(int i = 0;i < size;i++)
		memcpy(&temp[i], &buf[sizeof(int)*i], sizeof(int));
		
	sem_wait(&sem_sys);
	close(new_sd);
	sem_post(&sem_sys);

	merge(temp, size);

	delete[] temp;
	
	pthread_exit(0);
	return 0;
	
}

void main(int argc, char** argv)
{
	sem_init(&sem_sys, 0, 0);
	sem_post(&sem_sys);
	
	sem_init(&sem_local, 0, 0);
	sem_post(&sem_local);

	sem_init(&sem_sort, 0, 0);
	sem_post(&sem_sort);

	int n;
	int bytes_to_read;
	int port;
	int sd, new_sd;
	socklen_t client_len;

	struct sockaddr_in server, client;
	char *bp;
	char buf[BUFLEN];

	cout << "\nEnter the number of elements: ";	
	cin >> no_of_elements;

	elements = new int[no_of_elements];
	sorted_elements = new int[no_of_elements];
	
	cout << "Enter the elements: " << endl;

	for(int i = 0;i < no_of_elements;i++)
	{
		cout << i+1 << ": ";
		cin >> elements[i];
	}

	current_index = 0;

	port = SERVER_TCP_PORT;

	/* create a stream socket */

	if((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		cerr << "server: can't create a socket";
		exit(1);
	}

	/* Bind address to a socket */

	bzero((char*) &server, sizeof(struct sockaddr_in));
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	server.sin_addr.s_addr = htonl(INADDR_ANY);

	if(bind(sd, (struct sockaddr*) &server, sizeof(server)) == -1)
	{
		cerr << "can't bind name to socket";
		exit(1);
	}

	listen(sd, 5);

	while(1)
	{
		client_len = sizeof(client);

		if((new_sd = accept(sd, (struct sockaddr*) &client, &client_len)) == -1)
		{
			cerr << "server: can't accept client\n";
			exit(1);
		}

		if(fcntl(new_sd, F_SETFL, O_NONBLOCK) == -1)
		{
			cerr << "fcntl failed" << endl;
			exit(1);
		}

		pthread_t new_thread;
		struct thread_params t;
		t.new_sd = new_sd;
		t.current_index = current_index;

		pthread_create(&new_thread, 0, &distribute_problem ,&t);
		
		sem_wait(&sem_local);
		current_index = current_index + interval < no_of_elements ? current_index + interval : no_of_elements ;
		sem_post(&sem_local);

		if ( current_index == no_of_elements)
			break;
	}
	
	sem_wait(&sem_sys);
	close(sd);
	sem_post(&sem_sys);

	for(;;)
	{
		sem_wait(&sem_sys);
		if(done)
		{
			cout << "Sorted elements: ";
			for(int i = 0;i < no_of_elements;i++)
				cout << sorted_elements[i] << " ";
			cout << endl;
			
			delete[] elements;
			delete[] sorted_elements;
			
			sem_post(&sem_sys);
			break;
		}
		sem_post(&sem_sys);
	}
}