/* TCP Server */

#include <stdio.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#define SERVER_TCP_PORT  3000
#define BUFLEN 		 256
#define DELIMITER	'd'

struct thread_params
{
	int new_sd;
	int row;
	int col;
};

sem_t sem_sys;
sem_t sem_local;

int matrix1[10][10];
int matrix2[10][10];
int matrix3[10][10];

int row_size1, row_size2;
int col_size1, col_size2;

int no_of_calculated = 0;

void* distribute_problem(void *arg)
{
	struct thread_params *t = (struct thread_params*) arg;

	int new_sd = t->new_sd;
	int row = t->row;
	int col = t->col;

	/*** write the row and col to the client ***/

	char buf[BUFLEN];
	
	sem_wait(&sem_local);
	int r1 = row_size1;
	int r2 = row_size2;
	int c1 = col_size1;
	int c2 = col_size2;
	sem_post(&sem_local);

	/*** first copy row ***/
	sem_wait(&sem_local);
	memcpy(buf, &matrix1[row][0], c1 * sizeof(int));
	sem_post(&sem_local);

	/*** insert delimiter ***/
	buf[c1*sizeof(int)] = 'd';

	/*** copy cols ***/
	sem_wait(&sem_local);
	
	for(int i = 0;i < r2;i++)
		memcpy(&buf[c1*sizeof(int) + 1 + i*sizeof(int)], &matrix2[i][col], sizeof(int));
	
	sem_post(&sem_local);
	
	/*** copy delimiter ***/
	buf[c1*sizeof(int) + r2*sizeof(int) + 2] = 'd';

	int n;	
	for(;;)
	{
		sem_wait(&sem_sys);
		n = write(new_sd, buf, BUFLEN);
		sem_post(&sem_sys);
	
		if(n <= 0)
		{
			if(errno != EAGAIN)
			{
				cerr << "write failed" << endl;		
				pthread_exit(arg);
			}
		}
		else
			break;
		
	}

	/*** read the result ***/

	char *bp = buf;
	int bytes_to_read = BUFLEN;

	for(;;)
	{
		sem_wait(&sem_sys);
		n = read(new_sd, bp, bytes_to_read);
		sem_post(&sem_sys);

		if(n <= 0)
		{
			if(errno != EAGAIN)
			{
				cerr << "read failed" << endl;	
				pthread_exit(arg);
			}
			else
				continue;
		}

		bytes_to_read -= n;
		bp += n;

		if(!bytes_to_read)
			break;
				
	}


	int this_element;

	memcpy(&this_element, buf, sizeof(int));

	cout << "partial product: " << this_element << endl;		
	sem_wait(&sem_sys);
	close(new_sd);
	sem_post(&sem_sys);

	sem_wait(&sem_local);
	matrix3[row][col] = this_element;
	sem_post(&sem_local);

	sem_wait(&sem_local);
	no_of_calculated ++;
	sem_post(&sem_local);
	
	
	pthread_exit(0);
	return 0;
	
}

void main(int argc, char** argv)
{
	sem_init(&sem_sys, 0, 0);
	sem_post(&sem_sys);
	
	sem_init(&sem_local, 0, 0);
	sem_post(&sem_local);

	int n;
	int bytes_to_read;
	int port;
	int sd, new_sd;
	socklen_t client_len;

	struct sockaddr_in server, client;
	char *bp;
	char buf[BUFLEN];

	cout << "\nEnter the dimensions of matrix1" << endl;
	cout << "Rows: ";
	cin >> row_size1;
	cout << "\nCols: ";
	cin >> col_size1;

	cout << "\nEnter the dimensions of matrix2" << endl;
	cout << "Rows: ";
	cin >> row_size2;
	cout << "Cols: ";
	cin >> col_size2;

	if(col_size1 != row_size2)
	{
		cout << "Incompatible matrices" << endl;
		exit(1);
	}

	cout << "Enter the elements of matrix 1: " << endl;

	for(int i = 0;i < row_size1;i++)
		for(int j = 0;j < col_size1;j++)
			cin >> matrix1[i][j];

	cout << "Enter the elements of matrix 2: " << endl;

	for(int i = 0;i < row_size2;i++)
		for(int j =0;j < col_size2;j++)
			cin >> matrix2[i][j];

	port = SERVER_TCP_PORT;

	/* create a stream socket */

	if((sd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		cerr << "server: can't create a socket";
		exit(1);
	}

	/* Bind address to a socket */

	bzero((char*) &server, sizeof(struct sockaddr_in));
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	server.sin_addr.s_addr = htonl(INADDR_ANY);

	if(bind(sd, (struct sockaddr*) &server, sizeof(server)) == -1)
	{
		cerr << "can't bind name to socket";
		exit(1);
	}

	listen(sd, 5);

	int current_row = 0, current_col = 0;

	while(1)
	{
		client_len = sizeof(client);

		if((new_sd = accept(sd, (struct sockaddr*) &client, &client_len)) == -1)
		{
			cerr << "server: can't accept client\n";
			exit(1);
		}

		if(fcntl(new_sd, F_SETFL, O_NONBLOCK) == -1)
		{
			cerr << "fcntl failed" << endl;
			exit(1);
		}

		pthread_t new_thread;
		struct thread_params t;
		t.new_sd = new_sd;
		t.row = current_row;
		t.col = current_col;

		pthread_create(&new_thread, 0, &distribute_problem ,&t);
		
		sem_wait(&sem_local);

		if(current_col == col_size2 - 1)
		{
			current_col = 0;
			if(current_row + 1 == row_size1)
			{
				sem_post(&sem_local);
				break;
			}
			current_row++;
		}
		else
			current_col++;
			
		sem_post(&sem_local);
	}
	
	sem_wait(&sem_sys);
	close(sd);
	sem_post(&sem_sys);

	for(;;)
	{
		sem_wait(&sem_sys);
		if(no_of_calculated == row_size1 * col_size2)
		{
			cout << "Product of matrices: " << endl;
			for(int i = 0;i < row_size1;i++)
			{
				for(int j = 0;j < col_size2;j++)
					cout << matrix3[i][j] << " ";
				cout << endl;
			}		
				
			cout << endl;
			
			sem_post(&sem_sys);
			break;
		}
		sem_post(&sem_sys);
	}
}
